from nemspy.model.base import EntryType, ModelEntry


class HydrologicalModelEntry(ModelEntry):
    """
    abstraction of a generic hydrological model
    """

    entry_type = EntryType.HYDROLOGICAL

    def __init__(self, processors: int, **kwargs):
        super().__init__(processors, **kwargs)


class NationalWaterModelEntry(HydrologicalModelEntry):
    """
    National Water Model
    https://water.noaa.gov/about/nwm
    """

    name = 'nwm'

    def __init__(self, processors: int, **kwargs):
        super().__init__(processors, **kwargs)



from os import PathLike

from nemspy.model.base import EntryType, FileForcingEntry, ModelEntry


class WaveModelEntry(ModelEntry):
    """
    abstract implementation of a generic wave model
    """

    entry_type = EntryType.WAVE

    def __init__(self, processors: int, **kwargs):
        super().__init__(processors, **kwargs)


class WaveWatch3ForcingEntry(WaveModelEntry, FileForcingEntry):
    """
    file forcing entry for WaveWatch III data

    >>> wave_mesh = WaveWatch3ForcingEntry(filename='ww3.Constant.20151214_sxy_ike_date.nc', processors=1)
    """

    name = 'ww3data'

    def __init__(self, filename: PathLike = None, processors: int = None, **kwargs):
        if processors is None:
            processors = 1
        # Uses ww3data as name but the implementation is model agnostic
        WaveModelEntry.__init__(self, processors, **kwargs)
        FileForcingEntry.__init__(self, self.entry_type, filename)


class WaveWatch3Entry(WaveModelEntry):
    """
    WaveWatch III model
    https://polar.ncep.noaa.gov/waves/wavewatch/
    """

    name = 'ww3'

    def __init__(self, processors: int, **kwargs):
        super().__init__(processors, **kwargs)


class SWANEntry(WaveModelEntry):
    """
    SWAN wave model
    http://swanmodel.sourceforge.net/
    """

    name = 'swan'

    def __init__(self, processors: int, **kwargs):
        super().__init__(processors, **kwargs)



from abc import ABC, abstractmethod
from datetime import datetime, timedelta
from enum import Enum
import os
from os import PathLike
from pathlib import Path
import sys
from textwrap import indent
from typing import Iterator, List, Tuple, Union

if sys.version_info >= (3, 8):
    from importlib import metadata as importlib_metadata
else:
    import importlib_metadata

from nemspy.model.base import (
    AttributeEntry,
    ConnectionEntry,
    EntryType,
    FileForcingEntry,
    GridRemapMethod,
    INDENTATION,
    MediationEntry,
    MediatorEntry,
    ModelEntry,
    SequenceEntry,
    VerbosityOption,
)
from nemspy.utilities import create_symlink, LOGGER


class Earth(AttributeEntry):
    """
    multi-model coupling container representing the entire Earth system

    Only one of each model type can be assigned to the Earth system model at a time.
    """

    entry_title = 'EARTH'

    def __init__(self, **models):
        """
        :param atm: atmospheric wind model
        :param wav: oceanic wave model
        :param ocn: oceanic circulation model
        :param hyd: terrestrial water model
        :param med: model mediator
        """

        if 'Verbosity' not in models:
            models['Verbosity'] = VerbosityOption.OFF

        self.__models = {model_type: None for model_type in EntryType}

        attributes = {}
        for key, value in models.items():
            if key.upper() in {entry.name for entry in EntryType}:
                if isinstance(value, ModelEntry):
                    self[EntryType[key.upper()]] = value
            else:
                attributes[key] = value
        self.attributes = attributes

    @property
    def models(self):
        """
        list of models comprising the Earth system
        """

        return self.__models

    def __getitem__(self, model_type: EntryType) -> ModelEntry:
        return self.__models[model_type]

    def __setitem__(self, model_type: EntryType, model: ModelEntry):
        assert model_type == model.entry_type
        if self.__models[model_type] is not None:
            LOGGER.warning(
                f'overwriting existing "{model_type.name}" model: ' f'{repr(self[model_type])}'
            )
        self.__models[model_type] = model

    def __contains__(self, model_type: EntryType):
        return model_type in self.__models

    def __iter__(self) -> Iterator[Tuple[EntryType, ModelEntry]]:
        for model_type, model in self.models.items():
            yield model_type, model

    def __str__(self) -> str:
        attributes = [
            f'{attribute} = {value if not isinstance(value, Enum) else value.value}'
            for attribute, value in self.attributes.items()
        ]

        return '\n'.join(
            [
                f'{self.entry_title}_component_list: '
                f'{" ".join(model_type.value for model_type, model in self.models.items() if model is not None)}',
                f'{self.entry_title}_attributes::',
                indent('\n'.join(attributes), INDENTATION),
                '::',
            ]
        )

    def __repr__(self) -> str:
        models = [
            f'{model_type.name}={repr(model)}' for model_type, model in self.models.items()
        ]
        models += [f'{key}={value}' for key, value in self.attributes.items()]
        return (
            f'{self.__class__.__name__}({self.attributes["Verbosity"]}, {", ".join(models)})'
        )


class RunSequence(AttributeEntry, SequenceEntry):
    """
    multi-model container for model entries, defining the sequence in which they run within the modeled time loop

    NOTE: Currently, only one loop is supported. Nested loops will be implemented in a future version of NEMSpy.
    """

    entry_title = 'Run Sequence'

    def __init__(self, interval: timedelta, **kwargs):
        """
        :param interval: time interval to repeat the main loop in modeled time
        """

        self.interval = interval

        if 'Verbosity' not in kwargs:
            kwargs['Verbosity'] = VerbosityOption.OFF

        self.__models = {}
        attributes = {}
        for key, value in kwargs.items():
            model_types = [model_type.value for model_type in EntryType]
            if key.upper() in model_types and isinstance(value, ModelEntry):
                self.__models[EntryType(key.upper())] = value
            else:
                attributes[key] = value
        self.attributes = attributes

        self.__sequence = [
            model for model in self.models if model.entry_type != EntryType.MEDIATOR
        ]
        self.__link_models()

    def append(self, entry: SequenceEntry):
        """
        add a sequence entry
        """

        if isinstance(entry, ModelEntry):
            model_type = entry.entry_type
            if model_type in self.__models:
                del self.__models[model_type]
                self[entry.entry_type] = entry
        self.__sequence.append(entry)

    def extend(self, sequence: List[SequenceEntry]):
        """
        add several sequence entries
        """

        for entry in sequence:
            self.append(entry)

    @property
    def sequence(self) -> List[SequenceEntry]:
        """
        list of sequence entries in order, including model entries and connections / mediations
        """

        return self.__sequence

    @sequence.setter
    def sequence(self, sequence: List[SequenceEntry]):
        """
        set the sequence by passing a list of entries in order
        """

        sequence = list(sequence)
        if sequence != self.__sequence:
            mediator = self.mediator
            self.__models = {}
            if mediator is not None:
                self.mediator = mediator
            for entry in sequence:
                if isinstance(entry, ModelEntry):
                    model_type = entry.entry_type
                    if model_type in self.__models:
                        raise TypeError(
                            f'duplicate model type ' f'"{model_type.name}" in given sequence'
                        )
                    self.__models[model_type] = entry
            self.__link_models()
            self.__sequence = sequence

    def connect(
        self, source: EntryType, target: EntryType, method: GridRemapMethod = None, **kwargs,
    ):
        """
        assign a simple connection (not a mediation) between two model entries within the sequence
        """

        if method is None:
            method = GridRemapMethod.REDISTRIBUTE
        if EntryType.MEDIATOR in [source, target] and self.mediator is None:
            self.mediator = MediatorEntry(**kwargs)
        if source not in self.__models:
            raise KeyError(f'no {source.name} model in sequence')
        if target not in self.__models:
            raise KeyError(f'no {target.name} model in sequence')
        self.append(ConnectionEntry(self[source], self[target], method))

    @property
    def connections(self) -> List[Union[ConnectionEntry, MediationEntry]]:
        """
        list of all connections in the sequence
        """

        return [
            entry
            for entry in self.sequence
            if isinstance(entry, ConnectionEntry) or isinstance(entry, MediationEntry)
        ]

    @property
    def mediator(self) -> MediatorEntry:
        """
        shortcut property to the mediator entry
        """

        if EntryType.MEDIATOR in self:
            return self.__models[EntryType.MEDIATOR]
        else:
            return None

    @mediator.setter
    def mediator(self, mediator: MediatorEntry):
        """
        set the mediator entry (does not exist in the sequence by itself)
        """

        self[EntryType.MEDIATOR] = mediator

    def mediate(
        self,
        sources: List[EntryType] = None,
        functions: List[str] = None,
        targets: List[EntryType] = None,
        method: GridRemapMethod = None,
        processors: int = None,
        **attributes,
    ):
        """
        assign a mediation between two entries in the sequence
        """

        if 'name' not in attributes:
            attributes['name'] = 'mediator'
        if self.mediator is None:
            self.mediator = MediatorEntry(processors=processors, **attributes)
        else:
            self.mediator.attributes.update(attributes)
        if processors is not None:
            # increase mediation processor assignment if required
            if self.mediator.processors < processors:
                self.mediator.processors = processors

        if sources is not None:
            sources = [self[source] for source in sources]
        if targets is not None:
            targets = [self[target] for target in targets]

        self.append(MediationEntry(self.mediator, sources, functions, targets, method))

    @property
    def mediations(self) -> List[MediationEntry]:
        """
        list of all mediations in the sequence
        """

        return [entry for entry in self.sequence if isinstance(entry, MediationEntry)]

    @property
    def earth(self) -> Earth:
        """
        Earth system assigned to the sequence
        """

        return Earth(
            **{model.entry_type.name: model for model in self.models}, **self.attributes
        )

    @property
    def processors(self) -> int:
        """
        total number of processors assigned to sequence entries
        """

        return sum(model.processors for model in self.__models.values())

    def __link_models(self):
        """
        link entries and assign processors
        """

        models = self.models
        for model in models:
            if model.previous is not None:
                model.previous.next = None
            if model.next is not None:
                model.next = None
            model.start_processor = 0
        for model_index, model in enumerate(models):
            previous_model_index = model_index - 1
            if previous_model_index >= 0:
                model.previous = models[previous_model_index]

    def __setitem__(self, model_type: EntryType, model: ModelEntry):
        assert model_type == model.entry_type
        if model_type in self.__models:
            existing_model = self.__models[model_type]
            LOGGER.warning(
                f'overwriting {model_type.name} model ' f'"{existing_model}" with "{model}"'
            )
            self.__sequence.remove(self.__sequence.index(existing_model))
        self.__models[model_type] = model
        self.__link_models()

    def __getitem__(self, model_type: EntryType) -> ModelEntry:
        return self.__models[model_type]

    @property
    def models(self) -> List[ModelEntry]:
        """
        list of models in the run sequence
        """

        models = [
            model
            for model_type, model in self.__models.items()
            if model_type in self and model_type is not EntryType.MEDIATOR
        ]
        if self.mediator is not None:
            models.insert(0, self.mediator)
        return models

    def __iter__(self) -> Iterator[ModelEntry]:
        for model in self.models:
            yield model

    def __contains__(self, model_type: EntryType) -> bool:
        return model_type in self.__models

    def __len__(self) -> int:
        return len(self.sequence)

    @property
    def sequence_entry(self) -> str:
        return str(self)

    def __str__(self) -> str:
        block = '\n'.join(
            [
                f'@{self.interval / timedelta(seconds=1):.0f}',
                indent(
                    '\n'.join(entry.sequence_entry for entry in self.__sequence), INDENTATION
                ),
                '@',
            ]
        )
        return '\n'.join([f'runSeq::', indent(block, INDENTATION), '::'])

    def __repr__(self) -> str:
        models = [f'{model.entry_type.name.lower()}={repr(model)}' for model in self.models]
        return f'{self.__class__.__name__}({repr(self.interval)}, {", ".join(models)})'


class ConfigurationFile(ABC):
    """
    abstraction of a configuration file
    """

    name: str = NotImplementedError

    def __init__(self, sequence: RunSequence):
        """
        :param sequence: run sequence object containing models and order
        """

        self.sequence = sequence

    def __getitem__(self, entry_type: type) -> List[AttributeEntry]:
        return [entry for entry in self if isinstance(entry, entry_type)]

    @property
    def version_header(self) -> str:
        """
        comment header indicating filename and NEMSpy version
        """

        installed_distributions = importlib_metadata.distributions()
        for distribution in installed_distributions:
            if (
                distribution.metadata['Name'] is not None
                and distribution.metadata['Name'].lower() == 'nemspy'
            ):
                version = distribution.version
                break
        else:
            version = 'unknown'
        return f'# `{self.name}` generated with NEMSpy {version}'

    def write(
        self, filename: PathLike, overwrite: bool = False, include_version: bool = False
    ) -> Path:
        """
        write this configuration to file

        :param filename: path to file
        :param overwrite: overwrite an existing file
        :param include_version: include NEMSpy version information
        :returns: path to written file
        """

        if not isinstance(filename, Path):
            filename = Path(filename)
        ensure_directory(filename.parent)

        output = f'{self}\n'
        if include_version:
            output = f'{self.version_header}\n' f'{output}'

        if filename.is_dir():
            filename = filename / self.name
            LOGGER.warning(
                f'creating new file "{os.path.relpath(filename.resolve(), Path.cwd())}"'
            )

        if filename.exists():
            LOGGER.warning(
                f'{"overwriting" if overwrite else "skipping"} existing file "{os.path.relpath(filename.resolve(), Path.cwd())}"'
            )
        if not filename.exists() or overwrite:
            with open(filename, 'w', newline='\n') as output_file:
                output_file.write(output)

        return filename

    @abstractmethod
    def __str__(self) -> str:
        """
        :returns: string representation of configuration file
        """

        raise NotImplementedError

    def __repr__(self) -> str:
        return f'{self.__class__.__name__}({repr(self.sequence)})'


class NEMSConfigurationFile(ConfigurationFile):
    """
    ``nems.configure`` file, containing NEMS members, coupling connections, and run sequence information
    """

    name = 'nems.configure'

    @property
    def entries(self) -> List[AttributeEntry]:
        """
        list of attributed configuration entries, including the Earth system, models, and run sequence
        """

        return [self.sequence.earth, *self.sequence.models, self.sequence]

    def __iter__(self) -> Iterator[AttributeEntry]:
        for entry in self.entries:
            yield entry

    def __str__(self) -> str:
        return '\n'.join(f'# {entry.entry_title} #\n' f'{entry}\n' for entry in self).strip()


class FileForcingsFile(ConfigurationFile):
    """
    ``config.rc`` file, containing paths to forcing files
    """

    name = 'config.rc'

    @property
    def entries(self) -> List[FileForcingEntry]:
        """
        list of file forcing entries
        """

        return [entry for entry in self.sequence if isinstance(entry, FileForcingEntry)]

    def __iter__(self) -> Iterator[AttributeEntry]:
        for entry in self.entries:
            yield entry

    def __str__(self) -> str:
        return '\n'.join([FileForcingEntry.__str__(model_mesh) for model_mesh in self])


class ModelConfigurationFile(ConfigurationFile):
    """
    ``model_configure`` file, containing information on modeled start and end times, as well as ensemble information also aliased to ``atm_namelist.rc``
    """

    name = 'model_configure'

    def __init__(
        self,
        start_time: datetime,
        duration: timedelta,
        sequence: RunSequence,
        create_atm_namelist_rc: bool = True,
    ):
        """
        :param start_time: start time in model time
        :param duration: duration in model time
        :param sequence: run sequence containing models, connections, and order
        :param create_atm_namelist_rc: whether to create a symlink to ``atm_namelist.rc``
        """

        self.start_time = start_time
        self.duration = duration
        self.create_atm_namelist_rc = create_atm_namelist_rc
        super().__init__(sequence)

    def write(
        self, filename: PathLike, overwrite: bool = False, include_version: bool = False,
    ) -> Path:
        filename = super().write(filename, overwrite, include_version)
        if self.create_atm_namelist_rc:
            create_symlink(filename, filename.parent / 'atm_namelist.rc', relative=True)
        return filename

    def __str__(self) -> str:
        duration_hours = round(self.duration / timedelta(hours=1))
        return '\n'.join(
            [
                'total_member:            1',
                'print_esmf:              .true.',
                f'namelist:                {"atm_namelist.rc" if self.create_atm_namelist_rc else "model_configure"}',
                f'PE_MEMBER01:             {self.sequence.processors}',
                f'start_year:              {self.start_time.year}',
                f'start_month:             {self.start_time.month}',
                f'start_day:               {self.start_time.day}',
                f'start_hour:              {self.start_time.hour}',
                f'start_minute:            {self.start_time.minute}',
                f'start_second:            {self.start_time.second}',
                f'nhours_fcst:             {duration_hours:.0f}',
                'RUN_CONTINUE:            .false.',
                'ENS_SPS:                 .false.',
                # 'dt_atmos:                   @[DT_ATMOS]'
                # 'atm_coupling_interval_sec:  @[coupling_interval_fast_sec]'
                #
                # 'iatm: @[IATM]'
                # 'jatm: @[JATM]'
                #
                # 'cdate0: @[CDATE]'
                # 'nfhout: @[NFHOUT]'
                # 'filename_base: @[FILENAME_BASE]'
            ]
        )


def ensure_directory(directory: PathLike) -> Path:
    """
    ensure that a directory exists

    :param directory: directory path to ensure
    :returns: path to ensured directory
    """

    if not isinstance(directory, Path):
        directory = Path(directory)
    directory = directory.expanduser()
    if directory.is_file():
        directory = directory.parent
    if not directory.exists():
        directory.mkdir(parents=True, exist_ok=True)
    return directory



from abc import ABC, abstractmethod
from enum import Enum
from os import PathLike
from pathlib import PurePosixPath
from textwrap import indent
from typing import Dict, List

INDENTATION = '  '


class EntryType(Enum):
    """
    possible types of a configuration entry
    """

    ATMOSPHERIC = 'ATM'
    WAVE = 'WAV'
    OCEAN = 'OCN'
    HYDROLOGICAL = 'HYD'
    ICE = 'ICE'
    MEDIATOR = 'MED'


class VerbosityOption(Enum):
    """
    possible verbosity options for model entries
    """

    OFF = 'off'
    LOW = 'low'
    HIGH = 'high'
    MAX = 'max'


class GridRemapMethod(Enum):
    """
    possible remapping methods (interpolation / redistribution) for translating coupled data from one grid to another
    """

    REDISTRIBUTE = 'redist'
    BILINEAR = 'bilinear'
    PATH = 'patch'
    NEAREST_STOD = 'nearest_stod'
    NEAREST_DTOS = 'nearest_dtos'
    CONSERVE = 'conserve'


class FileForcingEntry(ABC):
    """
    abstraction of a forcing entry in ``config.rc``, defining the file path to a forcing file
    """

    def __init__(self, entry_type: EntryType, filename: PathLike = None):
        """
        :param entry_type: type of file forcing (i.e. ``ATM``, ``ICE``, etc.)
        :param filename: path to file
        """

        if filename is not None and not isinstance(filename, PurePosixPath):
            filename = PurePosixPath(filename)

        self.mesh_type = entry_type
        self.filename = filename

    def __str__(self) -> str:
        """
        string representation of the forcing entry in ``config.rc``
        """

        if self.filename is not None:
            directory = self.filename.parent.as_posix()
            name = self.filename.name
        else:
            directory = ''
            name = ''

        return (
            f' {self.mesh_type.value.lower()}_dir: {directory}\n'
            f' {self.mesh_type.value.lower()}_nam: {name}'
        )


class ConfigurationEntry(ABC):
    """
    abstraction of an entry in a configuration file
    """

    @classmethod
    @abstractmethod
    def from_string(cls, string: str, **kwargs) -> 'ConfigurationEntry':
        """
        parse entry information from a configuration entry string
        """

        raise NotImplementedError


class AttributeEntry(ABC):
    """
    abstraction of a configuration entry in ``nems.configure``
    """

    entry_title: str = NotImplementedError
    __attributes: Dict[str, str] = NotImplementedError

    @abstractmethod
    def __str__(self) -> str:
        """
        string representation of the configuration entry in ``nems.configure``
        """

        raise NotImplementedError

    @property
    def attributes(self) -> Dict[str, str]:
        """
        attributes defining options for the configuration entry, such as verbosity
        """

        attributes = {}
        for attribute, value in self.__attributes.items():
            if isinstance(value, Enum):
                value = value.value
            elif isinstance(value, bool):
                value = f'{value}'.lower()
            attributes[attribute] = value
        return attributes

    @attributes.setter
    def attributes(self, attributes: Dict[str, str]):
        self.__attributes = attributes


class SequenceEntry(ABC):
    """
    abstraction of an entry within the run sequence in ``nems.configure``
    """

    @property
    @abstractmethod
    def sequence_entry(self) -> str:
        """
        string representation of the sequence entry in ``nems.configure``
        """

        raise NotImplementedError

    def __str__(self) -> str:
        return self.sequence_entry


class ModelEntry(AttributeEntry, SequenceEntry, ConfigurationEntry):
    """
    abstraction of a generic model implementing NEMS / NUOPC coupling
    """

    entry_type: EntryType
    name: str

    def __init__(self, processors: int, **attributes):
        """
        The model entry is represented in two places in ``nems.configure``: once as a configuration entry with attributes, and once in the run sequence.
        The specific processors assigned to the model are defined by start and stop indices, which are determined on configuration write from the stop index of the previous entry in the run sequence.

        :param processors: number of processors to assign to this model
        """

        self.__processors = processors

        self.__start_processor = None

        self.__previous = None
        self.__next = None

        if 'Verbosity' not in attributes:
            attributes['Verbosity'] = VerbosityOption.OFF

        # http://www.earthsystemmodeling.org/esmf_releases/last_built/NUOPC_refdoc/node3.html#SECTION00033000000000000000
        self.attributes = attributes

    @property
    def entry_title(self) -> str:
        return str(self.entry_type.value) if self.entry_type is not None else None

    @property
    def processors(self) -> int:
        """
        the number of processors assigned to this model
        """

        return self.__processors

    @processors.setter
    def processors(self, processors: int):
        """
        set the number of processors (this also updates all subsequent entries in the run sequence)
        """

        if processors != self.processors:
            self.__processors = processors
            # update following processors
            self.start_processor = self.start_processor

    @property
    def start_processor(self) -> int:
        """
        the first index in the processor series assigned to this model
        """

        return self.__start_processor

    @start_processor.setter
    def start_processor(self, index: int):
        """
        set the first index in the processor series (this also updates all subsequent entries in the run sequence)
        """

        self.__start_processor = index
        current_model = self.next
        while current_model is not None:
            if current_model.previous is not None:
                current_model.__start_processor = current_model.previous.end_processor + 1
            current_model = current_model.next

    @property
    def end_processor(self) -> int:
        """
        the last index in the processor series assigned to this model
        """

        if self.start_processor is not None:
            return self.start_processor + self.processors - 1
        else:
            return None

    @property
    def previous(self) -> 'ModelEntry':
        """
        the previous entry in the run sequence
        """

        return self.__previous

    @previous.setter
    def previous(self, previous: 'ModelEntry'):
        """
        assign the previous entry in the run sequence (this also updates the processor series for this and all subsequent entries in the run sequence)
        """

        if previous is None and self.previous is not None:
            self.previous.__next = None
        self.__previous = previous
        if self.previous is not None:
            self.previous.__next = self
            if self.previous.end_processor is not None:
                self.start_processor = self.previous.end_processor + 1
        else:
            self.start_processor = 0

    @property
    def next(self) -> 'ModelEntry':
        """
        the next entry in the run sequence
        """

        return self.__next

    @next.setter
    def next(self, next: 'ModelEntry'):
        """
        assign the next entry in the run sequence (this updates all subsequent entries in the run sequence)
        """

        if next is None and self.next is not None:
            self.next.__previous = None
        self.__next = next
        if self.next is not None:
            self.next.__previous = self

    @property
    def sequence_entry(self) -> str:
        return str(self.entry_type.value)

    def __str__(self) -> str:
        return '\n'.join(
            [
                f'{self.entry_type.value}_model:                      {self.name}',
                f'{self.entry_type.value}_petlist_bounds:             {self.start_processor} {self.end_processor}',
                f'{self.entry_type.value}_attributes::',
                indent(
                    '\n'.join(
                        [
                            f'{attribute} = {value}'
                            for attribute, value in self.attributes.items()
                        ]
                    ),
                    INDENTATION,
                ),
                '::',
            ]
        )

    def __repr__(self) -> str:
        kwargs = [f'{key}={value}' for key, value in self.attributes.items()]
        return f'{self.__class__.__name__}({repr(self.name)}, {self.entry_type}, {self.processors}, {", ".join(kwargs)})'

    @classmethod
    def from_string(cls, string: str, **kwargs) -> 'ModelEntry':
        lines = string.splitlines()

        parsed_model_type, parsed_name = (value.strip() for value in lines[0].split('_model:'))
        parsed_model_type = EntryType(parsed_model_type)

        if hasattr(cls, 'model_type'):
            assert parsed_model_type == cls.entry_type
        if hasattr(cls, 'name'):
            assert parsed_name == cls.name

        start_processor, end_processor = [
            int(entry) for entry in lines[1].split('_petlist_bounds:')[-1].strip().split()
        ]

        attributes = {}
        for attribute_line in lines[3:-1]:
            key, value = (value.strip() for value in attribute_line.split('='))
            attributes[key] = value

        instance = cls(processors=end_processor + 1 - start_processor, **attributes, **kwargs,)

        if not hasattr(cls, 'model_type'):
            instance.entry_type = parsed_model_type
        if not hasattr(cls, 'name'):
            instance.name = parsed_name

        return instance


class ConnectionEntry(SequenceEntry, ConfigurationEntry):
    """
    a connection entry in ``nems.configure`` representing a simple coupling between two model entries
    """

    def __init__(self, source: ModelEntry, target: ModelEntry, method: GridRemapMethod = None):
        """
        :param source: source model entry
        :param target: target model entry
        :param method: remapping method with which to translate between differing grids (use ``redist`` for the same grid)
        """

        self.source = source
        self.target = target
        self.method = method if method is not None else GridRemapMethod.BILINEAR

    @property
    def models(self) -> List[ModelEntry]:
        """
        the source and target models in the coupling
        """

        return [self.source, self.target]

    @property
    def sequence_entry(self) -> str:
        return (
            f'{self.source.entry_type.value} -> {self.target.entry_type.value}'.ljust(13)
            + f':remapMethod={self.method.value}'
        )

    def __repr__(self) -> str:
        return f'{self.__class__.__name__}({repr(self.source)}, {repr(self.target)}, {repr(self.method)})'

    @classmethod
    def from_string(cls, string: str, **kwargs) -> 'ConnectionEntry':
        method = None
        try:
            source, target = (entry.strip() for entry in string.split('->', 1))
        except ValueError:
            raise ValueError(
                'connection entry should be formatted as `SRC -> DST   :remapMethod=METHOD`'
            )

        if ':' in target:
            target, parsed_method = (entry.strip() for entry in target.split(':', 1))
            if method is None and len(parsed_method) > 0:
                method = GridRemapMethod(parsed_method.split('=')[-1])

        source_model = ModelEntry(None)
        target_model = ModelEntry(None)

        source_model.name = source
        source_model.entry_type = EntryType(source)
        target_model.name = target
        target_model.entry_type = EntryType(target)

        return cls(source=source_model, target=target_model, method=method)


class MediatorEntry(ModelEntry):
    """
    a special entry in ``nems.configure`` representing a coupler between two model entries with a dedicated coupling function
    """

    entry_type = EntryType.MEDIATOR
    name = 'implicit'

    def __init__(self, processors: int = None, **attributes):
        """
        :param processors: number of processors to assign to this mediator
        """

        if processors is None:
            processors = 1
        super().__init__(processors, **attributes)


class MediationFunctionEntry(SequenceEntry, ConfigurationEntry):
    """
    the dedicated function of a mediation entry, applied to the coupling between two model entries in ``nems.configure``
    """

    def __init__(self, name: str, mediator: MediatorEntry):
        """
        :param name: name of function
        :param mediator: mediator entry to host this function
        """

        self.name = name
        self.mediator = mediator

    @property
    def sequence_entry(self) -> str:
        return f'{self.mediator.entry_type.value} {self.name}'

    def __repr__(self) -> str:
        return f'{self.__class__.__name__}({repr(self.name)}, {repr(self.mediator)})'

    @classmethod
    def from_string(cls, string: str, **kwargs) -> 'MediationFunctionEntry':
        """
        parse entry information from a configuration entry string
        """

        mediator_name, function_name = string.split()
        return cls(name=function_name, mediator=MediatorEntry(mediator_name))


class MediationEntry(SequenceEntry, ConfigurationEntry):
    """
    an application of a mediator between model entries, with a dedicated coupling function
    """

    def __init__(
        self,
        mediator: MediatorEntry,
        sources: List[ModelEntry] = None,
        functions: List[str] = None,
        targets: List[ModelEntry] = None,
        method: GridRemapMethod = None,
    ):
        """
        A Mediation can include multiple connections; for instance, an ICE model might connect to the mediation function, then the mediation function connects to the OCN model.

        :param mediator: mediator entry to host this mediation
        :param sources: source models
        :param functions: mediation functions to use
        :param targets: target models
        :param method: remapping method with which to translate between differing grids (use ``redist`` for the same grid)
        """

        if functions is None:
            functions = []

        self.mediator = mediator
        self.functions = [
            MediationFunctionEntry(mediation_function, self.mediator)
            for mediation_function in functions
        ]

        self.sources = sources
        self.targets = targets
        self.method = method

    @property
    def source_connections(self) -> List[ConnectionEntry]:
        """
        list of connections between the source(s) and the mediator
        """

        source_connections = []
        if self.sources is not None:
            for index, source in enumerate(self.sources):
                source_connections.append(ConnectionEntry(source, self.mediator, self.method))
        return source_connections

    @property
    def target_connections(self) -> List[ConnectionEntry]:
        """
        list of connections between the mediator and the target(s)
        """

        target_connections = []
        if self.targets is not None:
            for index, target in enumerate(self.targets):
                target_connections.append(ConnectionEntry(self.mediator, target, self.method))
        return target_connections

    @property
    def models(self) -> List[ModelEntry]:
        """
        list of model entries involved in the mediation
        """

        return [
            *(connection.source for connection in self.source_connections),
            self.mediator,
            *(connection.target for connection in self.target_connections),
        ]

    @property
    def sequence_entry(self) -> str:
        return '\n'.join(
            (
                *(
                    source_connection.sequence_entry
                    for source_connection in self.source_connections
                ),
                *(mediation_function.sequence_entry for mediation_function in self.functions),
                *(
                    target_connection.sequence_entry
                    for target_connection in self.target_connections
                ),
            )
        )

    def __repr__(self) -> str:
        return f'{self.__class__.__name__}({repr(self.mediator)}, {repr(self.sources)}, {repr(self.functions)}, {repr(self.targets)}, {repr(self.method)})'

    @classmethod
    def from_string(cls, string: str, **kwargs) -> 'MediationEntry':
        lines = string.strip().splitlines()

        connections = []
        functions = []
        for line in lines:
            if len(line) > 0:
                if '->' in line:
                    parts = [entry.strip() for entry in line.split('->')]
                    method = None
                    if ':' in parts[-1]:
                        parts[-1], method = [
                            entry.strip() for entry in parts[-1].split(':remapMethod=')
                        ]
                    if len(parts) == 2:
                        connection = ConnectionEntry.from_string(line)
                    else:
                        connection = ConnectionEntry(
                            source=parts[0], target=parts[-1], method=method
                        )
                    connections.append(connection)
                else:
                    functions.append(MediationFunctionEntry.from_string(line))

        mediator = None
        function_names = []
        for function in functions:
            if mediator is None:
                mediator = function.mediator
            function_names.append(function.name)

        method = None
        sources = []
        targets = []
        for connection in connections:
            if method is None:
                method = connection.method
            if connection.source.name != mediator.name:
                sources.append(connection.source)
            if connection.target.name != mediator.name:
                targets.append(connection.target)

        return cls(
            mediator=mediator,
            sources=sources,
            functions=function_names,
            targets=targets,
            method=method,
        )



from os import PathLike

from nemspy.model.base import EntryType, FileForcingEntry, ModelEntry


class IceModelEntry(ModelEntry):
    """
    abstraction of a generic ice model
    """

    entry_type = EntryType.ICE

    def __init__(self, processors: int, **kwargs):
        super().__init__(processors, **kwargs)


class IceForcingEntry(IceModelEntry, FileForcingEntry):
    """
    file forcing entry of an ice model
    """

    name = 'icemesh'

    def __init__(self, filename: PathLike = None, processors: int = None, **kwargs):
        if processors is None:
            processors = 1
        # Uses ww3data as name but the implementation is model agnostic
        IceModelEntry.__init__(self, processors, **kwargs)
        FileForcingEntry.__init__(self, self.entry_type, filename)
